from ..secrets.iring import * 
from morebs2.search_space_iterator import * 
from morebs2.matrix_methods import point_in_bounds
from copy import deepcopy 

DEFAULT_HOP_SIZE_RANGE = [2,9]
DEFAULT_BOUND_LENGTH_RANGE = [1,4] 

#------------------------ methods to check bounds 
def bounds_cover_actual_sec_vec(sec,bounds): 
    return point_in_bounds(bounds,sec.seq) 

def bounds_cover_one_optima_point_of_sec(sec,bounds): 
    opt_mat = sec.optima_points() 
    
    for o in opt_mat: 
        if point_in_bounds(bounds,o): return True 
    return False 

#--------------------------------------------------

def prng__search_space_bounds_for_vector(vec,hop_size,bound_length,prng=None):

    assert bound_length >= 1.0 

    # case: prng is None 
    if type(prng) == type(None): 
        bounds = np.array([vec,vec + bound_length])
        return bounds.T 

    X = bound_length / hop_size

    def hop_from_x(x,num_hops,is_back_hop:int):
        assert type(x) in {int,float,np.float32,np.float64} 

        X_ = X if not is_back_hop else -X 

        for _ in range(num_hops):
            x += X_ 
            #x = round(x,5) 
        return round(x,5) 

    # case: calculate a bound such that iterating over it using a search 
    #       space iterator yields the point `vec`. 
    #       bound is calculated using prng 
    V = [] 
    for i in range(len(vec)): 
        left_hops = int(prng()) % hop_size
        right_hops = hop_size - left_hops 

        v0 = hop_from_x(vec[i],left_hops,True) 
        v1 = hop_from_x(vec[i],right_hops,False) 
        V.append([v0,v1]) 
    return np.array(V)

def invalid_search_space_bounds_for_vector(vec,bound_length,prng=None): 

    assert bound_length >= 1.0 
    if type(prng) == type(None): 
        start = vec + bound_length
        end = start + end 
        return np.array([start,end]).T 

    # try five times with the `prng` to output an invalid bounds 
    t = 5 
    stat = False 
    bounds = None 
    while t > 0 and not stat: 
        start = np.array([prng() for _ in range(len(vec))]) 
        end = start + bound_length 
        bounds = np.array([start,end]).T 

        if point_in_bounds(bounds,vec): 
            t -= 1 
        else: 
            stat = True 

    if stat: 
        return bounds 
    return invalid_search_space_bounds_for_vector(vec,bound_length,None) 

def SearchSpaceIterator_for_bounds(bounds,hop_size): 
    startPoint = np.copy(bounds[:,0])
    columnOrder = [i for i in range(bounds.shape[0])]  
    cycleOn = False   
    cycleIs = 0 
    ssi = SearchSpaceIterator(bounds, startPoint, columnOrder, hop_size,cycleOn, cycleIs)
    return ssi 

"""
return: 
- bound for optima point, hop size for searching over bound, corresponding Pr. value for optima point. 
"""
# NOTE: `is_actual_sec_vec` may clash with `optima_point_index`.
def prng_leak_Secret(sec,prng=None,is_actual_sec_vec:bool=True,is_valid_bounds:bool=True,optima_point_index:int=None,\
    valid_one_shot_kill:bool=False,hop_size_range=DEFAULT_HOP_SIZE_RANGE):
    assert type(sec) == Sec 
    assert type(is_actual_sec_vec) == bool 
    assert type(is_valid_bounds) == bool 

    if type(prng) == type(None): 
        prng = default_std_Python_prng() 
 
    # fetch the target optima point 
    opt_mat = sec.optima_points()
    if is_actual_sec_vec: 
        seq = deepcopy(sec.seq) 
    elif type(optima_point_index) != type(None): 
        try:
            seq = opt_mat[optima_point_index] 
        except:
            raise ValueError("invalid optima point index")
    else:
        candidates = set([_ for _ in range(opt_mat.shape[0])]) - {sec.seq_index} 
        candidates = sorted(candidates) 
        index = int(prng()) % len(candidates) 
        
        index = candidates[index]
        seq = opt_mat[index] 

    # fetch the corresponding Pr. value for optima point 
    seq_str = vector_to_string(seq,float)
    pr_value = sec.opm[seq_str]

    # mask the optima point with a bound 
    if is_valid_bounds:
        hop_size = modulo_in_range(int(prng()),hop_size_range) 
        bound_length = modulo_in_range(int(prng()),DEFAULT_BOUND_LENGTH_RANGE)
        if not valid_one_shot_kill: 
            bounds = prng__search_space_bounds_for_vector(seq,hop_size,bound_length,prng)
        else: 
            bounds = np.array([deepcopy(seq),seq+hop_size]).T 
    else:
        hop_size = 4 
        bound_length = modulo_in_range(int(prng()),DEFAULT_BOUND_LENGTH_RANGE) 
        bounds = invalid_search_space_bounds_for_vector(seq,bound_length,prng)
    return bounds,hop_size,pr_value 

"""
return:
- dict, sec index -> (optima point index,bounds,hop_size,pr_value)
"""
def prng_leak_IsoRing_into_dict(ir:IsoRing,prng,actual_sec_vec_ratio=1.0,ratio_of_dim_covered=1.0,valid_bounds_ratio=1.0,\
    prioritize_actual_Sec:bool=True,valid_one_shot_kill_ratio=0.0,hop_size_range=DEFAULT_HOP_SIZE_RANGE): 

    def prg_(): return int(prng()) 

    num_sec = len(ir.sec_list) 
    num_dim_covered = ceil(num_sec * ratio_of_dim_covered)  

    # order of secrets 
    ilist = [_ for _ in range(num_sec)] 
    ilist = prg_seqsort(ilist,prg_) 

    # number of actual sec vec leaks for secrets 
    num_actual_sv_leaks = int(ceil(num_dim_covered * actual_sec_vec_ratio))

    # number of valid bounds
    num_valid_bounds = int(ceil(num_dim_covered * valid_bounds_ratio))

    # number of valid one-shot kills 
    num_valid_one_shot_kills = int(ceil(num_dim_covered * valid_one_shot_kill_ratio))

    if prioritize_actual_Sec: 
        i = ilist.index(ir.actual_sec_index) 
        j = ilist.pop(i) 
        ilist.insert(0,j) 
    ilist = ilist[:num_dim_covered]
    
    D = {}
    for i in ilist:
        s = ir.sec_list[i] 
        
        if num_actual_sv_leaks > 0: 
            is_actual_sec_vec = True 
            num_actual_sv_leaks -= 1 
        else: 
            is_actual_sec_vec = False 

        if num_valid_bounds > 0: 
            is_valid_bounds = True 
            num_valid_bounds -= 1 
        else: 
            is_valid_bounds = False 

        if num_valid_one_shot_kills > 0:
            valid_one_shot_kill = True 
            num_valid_one_shot_kills -= 1 
        else: 
            valid_one_shot_kill = False 
        
        seq_index = s.seq_index()
        if is_actual_sec_vec: 
            optima_point_index = seq_index
        else: 
            ix = [_ for _ in range(len(s.opm))]
            ix2 = ix.index(seq_index)
            ix.pop(ix2)

            if len(ix) != 0: 
                ix3 = prg_() % len(ix)
                optima_point_index = ix.pop(ix3) 
            else: 
                optima_point_index = 0 
        is_actual_sec_vec = False 

        bounds,hop_size,pr_value = prng_leak_Secret(s,prng=prng,is_actual_sec_vec=is_actual_sec_vec,\
            is_valid_bounds=is_valid_bounds,optima_point_index=optima_point_index,\
            valid_one_shot_kill=valid_one_shot_kill,hop_size_range=hop_size_range) 

        D[i] = (optima_point_index,bounds,hop_size,pr_value) 
    return D 


from .data_load import * 
import sys 

original_stdout = sys.stdout

def bool_prompt(prompt_string): 

    ask = input(prompt_string) 

    ask = ask.strip() 

    if ask == "0": return False 
    if ask == "1": return True 

    print("[!!] invalid input. try again.")
    return bool_prompt(prompt_string)

Q_NEG = "\t0 is NO, 1 is YES" 
Q0 = "[?] allow inaccuracies: "
Q1 = "[?] allow incomplete info: "
Q2 = "[?] allow wrong order of cracking: " 
Q3 = "[?] allow one shot kill: " 
Q4 = "[?] load Python standard random state: " 
Q5 = "[?] save Python standard random state: " 
Q6 = "[?] enter in a positive number (leave blank for infinity):  "
Q7 = "[?] enter in range for hop size (leave blank for default [2,9], min is 3,max is 9):  " 


def intro_message(): 
    print("This is an interface to simulate the security strength of an arbitrary\n")
    print("Isomorphic Ringed Chain against a third-party's brute force attempts.\n")
    print("Simulation is not completely applicable to real-world data security threats.\n")
    print("Use at your own discretion.")
    print("-------------------------------------------------------------------------------")

def prompt_loadORsave_std_Python_random_state(is_load:bool):
    if is_load: 
        S,F = Q4,load_std_Python_random_state
    else: 
        S,F = Q5,save_std_Python_random_state

    stat = bool_prompt(S)
    if not stat: return False 

    fp = input("[?] enter in filepath:  ")
    fp = fp.strip()
    return F(fp)

def prompt_vec_filepath(): 
    print("\t[!] secrets are vectors.")
    print("\t[!] vectors must be present on every line (50 vectors,50 lines max)") 
    print("\t[!] vectors must range in length between 1 through 5.")
    print("\t[!] EX:  101,110,100,111\n") 
    fp = input("[?] enter in filepath of vectors:  ")
    fp = fp.strip()

    try: 
        irc = load_vector_file_into_IsoRingedChain(fp,prng=default_std_Python_prng())
        return irc 
    except: 
        print("\t[!] invalid filepath OR erroneous file. try again.")
        return prompt_vec_filepath()

def prompt_hop_size(): 
    H = input(Q7)

    if H.strip() == "": return 9 

    try: 
        i = int(H) 
        assert 2 < i <= 9 
        return i 
    except: 
        print("invalid hop size. try again.")
        return prompt_hop_size()

def prompt_BackgroundInfo(irc): 
    print(Q_NEG)    
    q0 = bool_prompt(Q0) 
    q1 = bool_prompt(Q1) 
    q2 = bool_prompt(Q2) 
    q3 = bool_prompt(Q3) 
    q4 = prompt_hop_size() 
    prng = default_std_Python_prng() 
    return simulation_default_BackgroundInfo_for_IsoRingedChain(\
        irc,prng,q0,q1,q2,q3,q4)   

def prompt_float(): 
    T = input(Q6)
    if T.strip() == "": return float('inf')

    try: 
        x = float(T) 
        if x <= 0: 
            print("invalid number")
            return prompt_float()
        else: 
            return x 
    except: 
        print("invalid number")
        return prompt_float() 

def prompt_Cracker_energy(bi,irc): 

    F = prompt_float() 
    return instantiate_simulation_BruteForceEnv(bi,irc,F)
    
def redirect_print(): 

    # Save original stdout
    print_fp = input("[?] enter filepath for output:  ")

    try: 
        f = open(print_fp,"w")
        sys.stdout = f
        return f 
    except: 
        print("[:-(  invalid filepath.try again.")
     
    return redirect_print()

def ui_method(): 
    intro_message()
    print(Q_NEG)     
    stat0 = prompt_loadORsave_std_Python_random_state(True) 
    if not stat0: prompt_loadORsave_std_Python_random_state(False)
    
    irc = prompt_vec_filepath()
    bi = prompt_BackgroundInfo(irc)
    bfe = prompt_Cracker_energy(bi,irc) 

    fobj = redirect_print()
    while not bfe.is_finished(): 
        next(bfe) 
    
    bfe.crck.soln_synopsis() 

    print("[$] remaining Cracker energy: {}".format(bfe.crck.energy)) 
    fobj.close() 
    sys.stdout = original_stdout
    
    print("DONE!")
    return 

if __name__ == "__main__":

    ui_method()


from isoring.secrets.iring import * 
import unittest

def Sec_sample_X(): 
    singleton_range = [-100,100] 
    dimension = 4 
    num_optima = 10 
    prng = default_std_Python_prng(integer_seed=527,output_range=[0,1000],rounding_depth=3)
    s = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=5,set_actual_as_max_pr=True)
    return s 

def IsoRing_sample_X(feedback_function_type):
    s = Sec_sample_X()
    prng = default_std_Python_prng(integer_seed=302,output_range=[0,1000],rounding_depth=3)
    return IsoRing.generate_IsoRing_from_one_secret(s,prng,feedback_function_type,\
        num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
        sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE)

### lone file test 
"""
python3 -m tests.test_iring 
"""
###
class IsoRingClass(unittest.TestCase):

    def test__IsoRing__generate_IsoRing_from_one_secret__case1(self):

        ir0 = IsoRing_sample_X(0)
        ir1 = IsoRing_sample_X(1) 
        s = Sec_sample_X() 

        V = ir0.actual_sec_vec()
        V2 = ir1.actual_sec_vec()

        assert (V==V2).all()
        assert (V == s.seq).all() 

        S = ir0.iso_repr()
        S2 = ir1.iso_repr() 

        assert (S.seq == S2.seq).all() 
        assert (S.seq == s.seq).all() 
        assert (ir0.sec_list[ir0.actual_sec_index].seq == S.seq).all() 
        assert (ir1.sec_list[ir1.actual_sec_index].seq == S.seq).all() 
        assert ir0.actual_sec_index != 0 

    def test__IsoRing__provide_feedback_distance_vec(self): 
        ir0 = IsoRing_sample_X(0)
        ir1 = IsoRing_sample_X(1) 

        Q = np.array([ 77.685,  30.222,  87.129, -23.642])
        Q2 = np.array([ 77.685,  31.222,  87.129, -20.642])
        s00 = ir0.provide_feedback_distance_vec(Q)
        s01 = ir0.provide_feedback_distance_vec(Q2) 
        assert s00[8] == 0.
        assert 4. > s01[8] > 3.

        s10 = ir1.provide_feedback_distance_vec(Q)
        s11 = ir1.provide_feedback_distance_vec(Q2) 

        assert 820 > s10[8] > 800 
        assert 505 > s11[8] > 500

if __name__ == '__main__':
    unittest.main()



from isoring.brute_forcer.background_info import * 
import unittest

def IsoRingedChainANDprng_sample_T(): 
    prng = prg__LCG(-66,3,7,-3212) 
    prng2 = prg__LCG(16.36,13.23,74.1434,37177.05) 
    def prng3(): return prng() + prng2() 

    vec_lengths = [4,8,9,13] 
    vec_list = [] 
    for i in range(0,15): 
        qi = i % 4
        vec_list.append(one_vec(prng2,vec_lengths[qi],[-2000.,2000.])) 

    irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng3,\
        num_blooms_range=[DEFAULT_NUM_BLOOMS,DEFAULT_NUM_BLOOMS+1],codep_ratio=0.4,\
        shuffle_isorepr=False)

    return irc,prng3 



### lone file test 
"""
python3 -m tests.test_background_info  
"""
###
class BackgroundInfoClass(unittest.TestCase):

    def test__BackgroundInfo__extract_from_IsoRingedChain__case1(self):
        irc,prng3 = IsoRingedChainANDprng_sample_T() 

        # case: complete information 
        bi = BackgroundInfo.extract_from_IsoRingedChain(irc,prng3,actual_sec_vec_ratio_range=[1.,1.],\
            dim_covered_ratio_range=[1.,1.],valid_bounds_ratio_range=[1.,1.],prioritize_actual_Sec_ratio=1.,\
            shuffle_OOC_ratio=0.0,suspected_isoring_to_sec_idn_error_ratio=0.0)

        assert bi.order_of_cracking == irc.ooc 

        D = {k:v.actual_sec_index for k,v in irc.ir_dict.items()}

        assert bi.suspected_isoring_to_sec_idn == D 

    def test__BackgroundInfo__extract_from_IsoRingedChain__case2(self):
        irc,prng3 = IsoRingedChainANDprng_sample_T() 
        D = {k:v.actual_sec_index for k,v in irc.ir_dict.items()}

        bi2 = BackgroundInfo.extract_from_IsoRingedChain(irc,prng3,actual_sec_vec_ratio_range=[1.,1.],\
            dim_covered_ratio_range=[1.,1.],valid_bounds_ratio_range=[1.,1.],prioritize_actual_Sec_ratio=1.,\
            shuffle_OOC_ratio=0.5,suspected_isoring_to_sec_idn_error_ratio=0.5)

        assert bi2.order_of_cracking != irc.ooc 

        c = 0 
        for k,v in D.items(): 
            if bi2.suspected_isoring_to_sec_idn[k] != v: 
                c += 1 
        assert c == 8 

        bi3 = BackgroundInfo.extract_from_IsoRingedChain(irc,prng3,actual_sec_vec_ratio_range=[1.,1.],\
            dim_covered_ratio_range=[1.,1.],valid_bounds_ratio_range=[1.,1.],prioritize_actual_Sec_ratio=1.,\
            shuffle_OOC_ratio=0.5,suspected_isoring_to_sec_idn_error_ratio=0.5)

        assert bi3.order_of_cracking != irc.ooc 

        c = 0 
        for k,v in D.items(): 
            if bi3.suspected_isoring_to_sec_idn[k] != v: 
                c += 1 
        assert c == 8 

    """
    demonstrates `valid one shot kill` feature
    """
    def test__BackgroundInfo__extract_from_IsoRingedChain__case3(self): 
        irc,prng3 = IsoRingedChainANDprng_sample_T() 

        bi2 = BackgroundInfo.extract_from_IsoRingedChain(irc,prng3,actual_sec_vec_ratio_range=[1.,1.],\
            dim_covered_ratio_range=[1.,1.],valid_bounds_ratio_range=[1.,1.],prioritize_actual_Sec_ratio=1.,\
            shuffle_OOC_ratio=0.5,suspected_isoring_to_sec_idn_error_ratio=0.0,valid_one_shot_kill_ratio_range=[1.,1.]) 

        for k,v in bi2.info.items(): 
            ir = irc.fetch_IsoRing(k)
            ir.reset_iso_repr() 
            sec = ir.iso_repr()
            
            hs = v[ir.actual_sec_index]
            assert np.all(hs.suspected_subbound[:,0] == ir.actual_sec_vec()) 
            
        return

if __name__ == '__main__':
    unittest.main()



from isoring.secrets.secret import * 
import unittest

### lone file test 
"""
python3 -m tests.test_secret 
"""
###
class SecClass(unittest.TestCase):

    def test__Sec__generate_bare_instance__case1(self):
        singleton_range = [-10**6,10**6]
        dimension = 6 
        num_optima = 8 
        prng = prg__LCG(67,81,23,2121) 

        for _ in range(10): 
            s = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=0,set_actual_as_max_pr=False)
            ##print(s) 
            assert type(s) == Sec 
            assert len(s.opm) == 8 
            assert s.optima_points().shape == (8,6)
        
        assert (s.seq == np.array([-997994.0,-998687.0,-999674.0,-999023.0,-999317.0,-999800.0])).all()

    def test__Sec__generate_bare_instance__case2(self):

        singleton_range = [-100,100] 
        dimension = 5 
        num_optima = 5 
        prng = default_std_Python_prng(integer_seed=55,output_range=[0,1000],rounding_depth=3)
        s = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=0,set_actual_as_max_pr=True)

        q = vector_to_string(s.seq,float) 
        assert s.opm[q] == max(s.opm.values()) 

        fp = "pickled_sec_sample"
        s.pickle_thyself(fp)

        s2 = Sec.unpickle_thyself(fp) 
        assert np.all(s2.seq == s.seq )
        assert s.opm == s2.opm 
        assert s.ds == s2.ds and s.cds == s2.cds 



if __name__ == '__main__':
    unittest.main()



from isoring.brute_forcer.leak_functions import * 
import unittest

def SecANDprng_sample_X(): 
    prng = prg__LCG(66,3,7,3212) 
    sec = Sec.generate_bare_instance([-60.,1112.],5,6,prng,idn_tag=0,set_actual_as_max_pr=False)
    return sec,prng 

def IsoRingANDprng_sample_X(): 
    prng = prg__LCG(770,31,72,3212) 
    sec = Sec.generate_bare_instance([-60.,1112.],5,6,prng,idn_tag=0,set_actual_as_max_pr=False)
    ir = IsoRing.generate_IsoRing_from_one_secret(sec,prng,feedback_function_type=0)
    return ir,prng 


### lone file test 
"""
python3 -m tests.test_leak_functions
"""
###
class LeakFunctions(unittest.TestCase):

    def test__prng__search_space_bounds_for_vector__case1(self): 
        prng = prg__LCG(67,43,101,2112)

        vec = np.array([5.654,67.01,55.32])
        hop_size = 6 
        bound_length = 1.0 
        bounds = prng__search_space_bounds_for_vector(vec,hop_size,bound_length,prng=prng)

        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size) 
        while not ssi.reached_end(): 
            q = next(ssi) 
            if euclidean_point_distance(q,vec) < 0.01: 
                return 
        assert False

    def test__prng__search_space_bounds_for_vector__case2(self): 
        prng = prg__LCG(671,432,9,2112)

        vec = np.array([56.5114,67.11201,551.2132,712.3413])
        hop_size = 9 
        bound_length = 1.0 
        bounds = prng__search_space_bounds_for_vector(vec,hop_size,bound_length,prng=prng)

        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size) 
        while not ssi.reached_end(): 
            q = next(ssi) 
            if euclidean_point_distance(q,vec) < 0.01: 
                return 
        assert False

    def test__prng__search_space_bounds_for_vector__case3(self): 
        prng = prg__LCG(671,432,9,2112)

        vec = np.array([11.1,311.24141,56.5114,67.11201,551.2132,712.3413])
        hop_size = 3 
        bound_length = 1.0 
        bounds = prng__search_space_bounds_for_vector(vec,hop_size,bound_length,prng=prng)

        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size) 
        while not ssi.reached_end(): 
            q = next(ssi) 
            if euclidean_point_distance(q,vec) < 0.01: 
                return 
        assert False

    """
    correct target local optimum 
    """
    def test__prng_leak_Secret__case1(self): 
        sec,prng = SecANDprng_sample_X() 

        bounds,hop_size,pr_value = prng_leak_Secret(sec,prng,is_actual_sec_vec=True,is_valid_bounds=True)
        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size)
        i = 0 
        while True: 
            q = next(ssi) 
            i += 1 
            if euclidean_point_distance(q,sec.seq) < 0.05:
                break 

        assert i == 2208,"got {}".format(i)

    """
    correct target local optimum 
    """
    def test__prng_leak_Secret__case2(self): 
        sec,prng = SecANDprng_sample_X() 
        bounds,hop_size,pr_value = prng_leak_Secret(sec,prng,is_actual_sec_vec=True,is_valid_bounds=False)
        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size)
        stat0,stat1 = False,False
        while not ssi.reached_end(): 
            q = next(ssi) 
            if euclidean_point_distance(q,sec.seq) < 0.05:
                stat0 = True 
            
            q_ = vector_to_string(q,float) 
            if q_ in sec.opm: 
                stat1 = True 

        assert not stat0 and not stat1
        return

    """
    incorrect target local optimum 
    """
    def test__prng_leak_Secret__case3(self): 
        sec,prng = SecANDprng_sample_X() 
        bounds,hop_size,pr_value = prng_leak_Secret(sec,prng,is_actual_sec_vec=False,is_valid_bounds=True)
        ssi = SearchSpaceIterator_for_bounds(bounds,hop_size)

        stat0,stat1 = False,False
        while not ssi.reached_end(): 
            q = next(ssi) 
            if euclidean_point_distance(q,sec.seq) < 0.05:
                stat0 = True 
            
            q_ = vector_to_string(q,float) 
            if q_ in sec.opm: 
                stat1 = True 

        assert not stat0 and stat1 
        return

    def test__prng_leak_IsoRing_into_dict__case1(self): 
        ir,prng = IsoRingANDprng_sample_X()
        D = prng_leak_IsoRing_into_dict(ir,prng,actual_sec_vec_ratio=1.0,ratio_of_dim_covered=1.0,valid_bounds_ratio=1.0,\
            prioritize_actual_Sec=True)

        assert len(D) == len(ir.sec_list)

        i = 0
        for k,v in D.items(): 
            sec = ir.sec_list[k]
            i += int(bounds_cover_actual_sec_vec(sec,v[1])) 
        assert i == len(ir.sec_list) 

        D2 = prng_leak_IsoRing_into_dict(ir,prng,actual_sec_vec_ratio=1.0,ratio_of_dim_covered=10**-6,valid_bounds_ratio=1.0,\
            prioritize_actual_Sec=True)

        assert list(D2.keys())[0] == ir.actual_sec_index

        prng2 = prg__LCG(0,1,2,3)
        D3 = prng_leak_IsoRing_into_dict(ir,prng2,actual_sec_vec_ratio=1.0,ratio_of_dim_covered=10**-6,valid_bounds_ratio=1.0,\
            prioritize_actual_Sec=False)

        assert list(D3.keys())[0] != ir.actual_sec_index

    """
    incorrect target local optimum 
    """
    def test__prng_leak_IsoRing_into_dict__case2(self): 
        ir,prng = IsoRingANDprng_sample_X()
        
        # subcase: valid bounds 
        D4 = prng_leak_IsoRing_into_dict(ir,prng,actual_sec_vec_ratio=0.0,ratio_of_dim_covered=1.0,valid_bounds_ratio=1.0,\
            prioritize_actual_Sec=False)

        i,j = 0,0
        for k,v in D4.items(): 
            sec = ir.sec_list[k]
            i += int(bounds_cover_actual_sec_vec(sec,v[1]))
            j += int(bounds_cover_one_optima_point_of_sec(sec,v[1])) 
        assert i == 0 and j == 7 

        # subcase: invalid bounds 
        D5 = prng_leak_IsoRing_into_dict(ir,prng,actual_sec_vec_ratio=0.0,ratio_of_dim_covered=1.0,valid_bounds_ratio=0.0,\
            prioritize_actual_Sec=False)

        i,j = 0,0
        for k,v in D5.items(): 
            sec = ir.sec_list[k]
            i += int(bounds_cover_actual_sec_vec(sec,v[1]))
            j += int(bounds_cover_one_optima_point_of_sec(sec,v[1])) 
        assert i == j == 0 

if __name__ == '__main__':
    unittest.main()



from isoring.brute_forcer.crackling import * 
import unittest

def IsoRingANDprng_sample_Y(): 
    singleton_range = [-10.,10.] 
    dimension = 5 
    num_optima = 5 
    prng = prg__LCG(43.12,-312.3,455,4000)
    prng2 = prg__LCG(43.12,455,-312.3,-4000)
    def prng3(): return prng() + prng2() 

    sec = Sec.generate_bare_instance(singleton_range,dimension,num_optima,prng,idn_tag=0,set_actual_as_max_pr=True)

    ir = IsoRing.generate_IsoRing_from_one_secret(sec,prng,feedback_function_type=0,\
            num_blooms=3,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
            sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,\
            optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE)

    return ir,prng,prng2,prng3


### lone file test 
"""
python3 -m tests.test_crackling
"""
###
class CBridgeClass(unittest.TestCase):

    def test__CBridge__next__case1(self):
        ir,_,prng2,_ = IsoRingANDprng_sample_Y() 

        actual_sec_vec_ratio = 0.0 
        ratio_of_dim_covered = 10 ** -6 
        valid_bounds_ratio = 1.0 
        hs_dict = HypStruct.extract_from_IsoRing_into_HypStruct_dict(ir,prng2,actual_sec_vec_ratio,\
                        ratio_of_dim_covered,valid_bounds_ratio,True)

        hs = hs_dict[6] 

        # subcase: wrong answer, Crackling accepts as correct answer 
        cr = Crackling(ir.idn_tag(),ir.actual_sec_index,hs.opt_index) 
        cb = CBridge(cr,hs,ir)

        while not cb.terminated: 
            next(cb) 
        assert cr.num_attempts == 26 

        assert not np.any(cr.cracked_soln == ir.actual_sec_vec()) 
        assert cr.soln_pr 

        hs_ = deepcopy(hs) 
        hs_.probability_marker += 0.05 

        # subcase: wrong answer, Crackling does not accept as correct answer 
        cr2 = Crackling(ir.idn_tag(),ir.actual_sec_index,hs.opt_index)
        cb2 = CBridge(cr2,hs_,ir)
        while not cb2.terminated: 
            next(cb2) 

        assert cr2.num_attempts == cr.num_attempts 
        assert np.all(cr.cracked_soln == cr2.cracked_soln) 
        assert not cr2.soln_pr 

        return 

    def test__CBridge__next__case2(self):
        ir,_,_,_ = IsoRingANDprng_sample_Y() 

        prng = prg__LCG(-38,-31,45,-4000)
        actual_sec_vec_ratio = 1.0 
        ratio_of_dim_covered = 10 ** -6 
        valid_bounds_ratio = 1.0 
        hs_dict2 = HypStruct.extract_from_IsoRing_into_HypStruct_dict(ir,prng,actual_sec_vec_ratio,\
                        ratio_of_dim_covered,valid_bounds_ratio,False) 
        hs3 = hs_dict2[4] 

            ## subcase: wrong dimension 
        cr3 = Crackling(ir.idn_tag(),ir.current_sec_index,hs3.opt_index)
        cb3 = CBridge(cr3,hs3,ir)
        while not cb3.terminated: 
            next(cb3) 

        assert type(cr3.cracked_soln) == type(None) 
        assert type(cr3.soln_pr) == type(None) 
        assert cr3.num_attempts == 37 


            ## subcase: correct for repr 4 
        ir.set_iso_repr(4) 
        cr4 = Crackling(ir.idn_tag(),ir.current_sec_index,hs3.opt_index)
        cb4 = CBridge(cr4,hs3,ir)
        while not cb4.terminated: 
            next(cb4) 

        assert np.all(cr4.cracked_soln == np.array([ -4.6239 , -10.55495])) 
        assert cr4.num_attempts == 4, "got {}".format(cr4.num_attempts)
        assert cr4.soln_pr

            ## subcase: correct for repr4, <Crackling> does not accept as 
            ##          correct answer 
        hs3.probability_marker += 0.05 
        cr5 = Crackling(ir.idn_tag(),ir.current_sec_index,hs3.opt_index)
        cb5 = CBridge(cr5,hs3,ir)
        while not cb5.terminated: 
            next(cb5) 

        assert np.all(cr5.cracked_soln == np.array([ -4.6239 , -10.55495])) 
        assert cr5.num_attempts == 4, "got {}".format(cr5.num_attempts)
        assert not cr5.soln_pr
        return 

if __name__ == '__main__':
    unittest.main()



from isoring.secrets.big_secret import * 
import unittest

def IsoRing_list_sample_Q(): 
    prng = prg__LCG(66,3,7,3212) 

    ir_list = [] 
    for i in range(15): 
        sec = Sec.generate_bare_instance([-60.,1112.],5,6,prng,idn_tag=i,set_actual_as_max_pr=False)

        ir = IsoRing.generate_IsoRing_from_one_secret(sec,prng,0,\
            num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
            sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,
            optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE)
        ir_list.append(ir) 

    IsoRingedChain.prng__add_depANDcodep_to_IsoRingList(ir_list,prng,codep_ratio=0.3)
    return ir_list 

def IsoRingedChain_Accessories_sample_R():

    prng = prg__LCG(66,3,7,3212) 
    prng2 = prg__LCG(16.36,13.23,74.1434,37177.05) 

    vec_lengths = [4,8,9,13] 
    vec_list = [] 
    for i in range(0,15): 
        qi = i % 4
        vec_list.append(one_vec(prng2,vec_lengths[qi],[-2000.,2000.])) 

    return vec_list,prng,prng2 

### lone file test 
"""
python3 -m tests.test_big_secret  
"""
###
class IsoRingedChainClass(unittest.TestCase):

    def test__IsoRingedChain__prng__idns_to_order_of_depANDcodep__case1(self):
        prng = default_std_Python_prng(integer_seed=302,output_range=[0,1000],rounding_depth=3)
        X = [6,1,3,10,12,43,56,13,31,42,113,57,65,78,110,95,93,91,16]

        Y = IsoRingedChain.prng__idns_to_order_of_depANDcodep(X[:],prng,0)
        for y in Y: 
            assert len(y) == 1 

        Y2 = IsoRingedChain.prng__idns_to_order_of_depANDcodep(X[:],prng,0.5)

        s = 0 
        for y in Y2: 
            if len(y) == 1: continue 
            s += len(y) - 1 
        assert s == ceil(0.5 * (len(X) - 1)), "got {} wanted {}".format(s,ceil(0.5 * (len(X) - 1)))
        return 


    def test__IsoRingedChain__calculate_OOC_for_IsoRing_list__case1(self): 

        ir_list = IsoRing_list_sample_Q()
        ooc,stat = IsoRingedChain.calculate_OOC_for_IsoRing_list(ir_list)
        assert stat

        sol = [{0, 12}, {4}, {5}, {6}, {8, 3, 11, 7}, {13}, {1}, {2}, {9, 14}, {10}]
        assert ooc == sol 

    def test__IsoRingedChain__calculate_OOC_for_IsoRing_list__case2(self): 

        ir_list = IsoRing_list_sample_Q()


        ## case 1 
        irings = ir_list[:6] 

        for ir in irings: 
            ir.clear_depANDcodep_sets() 

        irings[0].assign_DC_set({1,2,3},set())
        irings[1].assign_DC_set({0,2,3},set())
        irings[2].assign_DC_set({3},set())


        ooc,stat = IsoRingedChain.calculate_OOC_for_IsoRing_list(irings)
        assert not stat 

        ## case 2 
        for ir in irings: 
            ir.clear_depANDcodep_sets() 

        irings[0].assign_DC_set(set(),{1,2})
        irings[1].assign_DC_set(set(),{1,2})
        irings[2].assign_DC_set(set(),{1,2}) 
        irings[3].assign_DC_set({1,2},{0}) 

        ooc,stat = IsoRingedChain.calculate_OOC_for_IsoRing_list(irings)
        assert not stat 

        ## case 3 
        for ir in irings: 
            ir.clear_depANDcodep_sets()

        irings[0].assign_DC_set({1},set())
        irings[1].assign_DC_set({2},set())
        irings[2].assign_DC_set({3},set())
        irings[3].assign_DC_set(set(),set())
        irings[4].assign_DC_set({1,2,5},set())
        irings[5].assign_DC_set({0},set())

        ooc,stat = IsoRingedChain.calculate_OOC_for_IsoRing_list(irings)
        assert not stat 
        return 

    def test__IsoRingedChain__list_of_vectors_to_IsoRingedChain__case1(self): 
        
        vec_list,prng,prng2 = IsoRingedChain_Accessories_sample_R()
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(\
            vec_list,prng2,num_blooms_range=[DEFAULT_NUM_BLOOMS,DEFAULT_NUM_BLOOMS+1],
            shuffle_isorepr=False)

        dsize = 0 
        csize = 0 
        for ir in irc.ir_dict.values(): 
            dsize += len(ir.dc_set(True))
            csize += len(ir.dc_set(False)) 

        assert dsize == 105 == csize + 105 
        return

    def test__IsoRingedChain__list_of_vectors_to_IsoRingedChain__case2(self): 
        vec_list,prng,prng2 = IsoRingedChain_Accessories_sample_R()
        irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng,num_blooms_range=[DEFAULT_NUM_BLOOMS,DEFAULT_NUM_BLOOMS+1],\
            codep_ratio=1.0,shuffle_isorepr=False)

        dsize = 0 
        csize = 0 
        for ir in irc.ir_dict.values(): 
            dsize += len(ir.dc_set(True))
            csize += len(ir.dc_set(False)) 

        assert dsize == 26
        assert csize == 158

        L = set([len(o) for o in irc.ooc])
        assert L == {2,13} 

    def test__IsoRingedChain__list_of_vectors_to_IsoRingedChain__case3(self):
        prng = prg__LCG(-66,3,7,-3212) 
        prng2 = prg__LCG(16.36,13.23,74.1434,37177.05) 
        def prng3(): return prng() + prng2() 

        vec_lengths = [4,8,9,13] 
        vec_list = [] 
        for i in range(0,15): 
            qi = i % 4
            vec_list.append(one_vec(prng2,vec_lengths[qi],[-2000.,2000.])) 

        try: 
            irc = IsoRingedChain.list_of_vectors_to_IsoRingedChain(vec_list,prng3,\
                num_blooms_range=[DEFAULT_NUM_BLOOMS,DEFAULT_NUM_BLOOMS+1],codep_ratio=1.0,\
                shuffle_isorepr=True) 
        except: 
            assert False 

if __name__ == '__main__':
    unittest.main()



from .secret import * 
from math import ceil 

DEFAULT_BLOOM_VECTOR_DIM_RANGE = [2,15] 
DEFAULT_NUM_BLOOMS = 6 
DEFAULT_BLOOM_MULTIPLIER_RANGE = [0.5,2.0]

"""
Used to calculate additional <Sec> instances given a base <Sec>. These additional <Sec>s are 
calculated with the aid of a pseudo-random number generator, `prng`, and accomodating parameters. 
"""
class BloomOfSecret: 

    def __init__(self,sec,prng,num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
        sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE): 

        assert type(sec) == Sec 
        assert is_valid_range(dim_range,True,False) and dim_range[0] > 0
        assert num_blooms < dim_range[1] - dim_range[0]
        assert is_valid_range(sec_vec_multiplier_range,False,True) and sec_vec_multiplier_range[0] > 0.  
        assert is_valid_range(optima_multiplier_range,False,True) and optima_multiplier_range[0] > 0.  

        self.base_sec = sec 
        self.all_sec = [self.base_sec] 
        self.num_blooms = num_blooms 
        self.svec_multiplier_range = sec_vec_multiplier_range
        self.prng = prng 
        self.dim_range = dim_range
        self.om_range = optima_multiplier_range
        
        available_dim = set([_ for _ in range(dim_range[0],dim_range[1])])
        available_dim -= {self.base_sec.dim()} 
        self.available_dim = sorted(available_dim)

    def __next__(self): 
        if len(self.all_sec) == self.num_blooms +1: return None 

        r0,r1 = min(self.all_sec[-1].seq),max(self.all_sec[-1].seq) 
        d = r1 - r0 
        median = (r1 - r0) / 2.0 
        omult = modulo_in_range(self.prng(),self.svec_multiplier_range)
        new_dist = (d * omult) / 2.0 
        r0_,r1_ = r0 - new_dist, r1 + new_dist 
        singleton_range = [r0_,r1_]
        
        di_index = int(self.prng()) % len(self.available_dim) 
        dimension = self.available_dim.pop(di_index) 

        omult = modulo_in_range(self.prng(),self.om_range) 
        num_optima = ceil(len(self.all_sec[-1].opm) * omult)

        set_actual_as_max_pr = bool(int(self.prng()) % 2) 

        S = Sec.generate_bare_instance(singleton_range,dimension,num_optima,self.prng,\
            idn_tag=self.base_sec.idn_tag,set_actual_as_max_pr=set_actual_as_max_pr)
        self.all_sec.append(S)
        return S


from .bloominstein import * 
from morebs2.numerical_generator import prg_seqsort
from morebs2.matrix_methods import euclidean_point_distance 

def prng_point_distance_funtion(prng): 

    def f(x0,x1):
        return euclidean_point_distance(x0,x1) + prng()   
    return f 

class IsoRing:

    """
    sec_list := list, every element is a <Sec> 
    feedback_function := function, F: (vector,vector) -> float 
    """
    def __init__(self,sec_list,feedback_function,actual_sec_index=0): 
        assert len(sec_list) > 0 
        assert len(sec_list) > actual_sec_index >= 0 
        idn_tag = set() 
        for s in sec_list: 
            assert type(s) == Sec 
            idn_tag |= {s.idn_tag}
        assert len(idn_tag) == 1
        self.sec_list = sec_list 
        self.feedback_function = feedback_function
        self.actual_sec_index = actual_sec_index
        self.current_sec_index = actual_sec_index

        self.cracked_sec_indices = [] 
        return

    #------------------------ functions to accept guesses (cracking attempts) from third-party 

    """
    return: 
    - optima point, optima point index, probability value 
    """
    def guess_equals_one_feedback(self,point):
        sec = self.iso_repr() 

        opt_pts = sec.optima_points()

        for (i,o) in enumerate(opt_pts):
            dist = None 
            stat = None 
            pr = None 
            try: 
                dist = euclidean_point_distance(o,point)
                stat = dist < 0.05 
            except: 
                return None,None,None

            if stat: 
                pr = self.provide_feedback_pr(vector_to_string(o,float)) 
                assert pr != -1 
                return o,i,pr 
        return None,None,None 

    """
    return: 
    - vector of feedback scores, length equal to number of optima points for 
                                iso_repr. 
    """
    def provide_feedback_distance_vec(self,i):
        s = self.iso_repr() 
        opt_points = s.optima_points() 

        V = [] 
        for o in opt_points: 
            try: 
                v = self.feedback_function(i,o) 
                V.append(v) 
            # case: different dim. 
            except:
                return None 
        return np.array(V)  
    
    def provide_feedback_pr(self,stringized_opt_point:str): 
        s = self.iso_repr() 
        if stringized_opt_point not in s.opm: return -1 
        return s.opm[stringized_opt_point]

    #----------------------------- repr functions 

    def idn_tag(self): 
        return self.sec_list[0].idn_tag 

    def set_iso_repr(self,i):
        assert 0 <= i < len(self.sec_list) 
        self.current_sec_index = i 

    def iso_repr(self): 
        return self.sec_list[self.current_sec_index]

    def reset_iso_repr(self): 
        self.current_sec_index = self.actual_sec_index

    def actual_sec_vec(self): 
        return self.sec_list[self.actual_sec_index].seq 

    def fetch_Sec(self,i): 
        assert type(i) == int 
        assert 0 <= i < len(self.sec_list)
        return self.sec_list[i] 

    #---------------------------- switch of repr functions, for use in cracking situations 

    """
    return:
    - ?any uncracked Sec remaining? 
    """
    def register_cracked_sec_index(self,prng): 
        # case: already cracked 
        if self.current_sec_index in self.cracked_sec_indices: 
            return 

        def prg_(): return int(prng())

        # case: register as cracked 
        self.cracked_sec_indices.append(self.current_sec_index) 

            # switch to another <Sec>
        self.switch_iso_repr(prng)
        return True 

    def switch_iso_repr(self,prng): 
        available = [_ for _ in range(len(self.sec_list)) if _ not in self.cracked_sec_indices]  
            # case: none left 
        if len(available) == 0: 
            return False 

        i = int(prng()) % len(available)
        i = available[i] 
        self.set_iso_repr(i) 

    #------------------- dep/codep functions 

    def dc_set(self,is_dep:bool=True):
        s = self.sec_list[0]
        return s.ds if is_dep else s.cds 

    def assign_DC_set(self,ds,cds):
        assert type(ds) == set 
        for s in self.sec_list:
            s.ds = ds
            s.cds = cds 

    def clear_depANDcodep_sets(self): 
        for s in self.sec_list:
            s.ds.clear()
            s.cds.clear() 

    """
    feedback_function_type := 0 for euclidean point distance, 1 for prng noise added. 
    """
    @staticmethod 
    def generate_IsoRing_from_one_secret(sec,prng,feedback_function_type,\
        num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
        sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE): 

        assert feedback_function_type in {0,1}

        bos = BloomOfSecret(sec,prng,num_blooms=DEFAULT_NUM_BLOOMS,dim_range=DEFAULT_BLOOM_VECTOR_DIM_RANGE,\
        sec_vec_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE,optima_multiplier_range=DEFAULT_BLOOM_MULTIPLIER_RANGE)

        while True: 
            if type(next(bos)) == type(None): break 

        def prg_(): 
            return int(prng())

        l = [i for i in range(bos.num_blooms + 1)]
        l = prg_seqsort(l,prg_)
        
        actual_sec_index = l.index(0)
        sec_list = [bos.all_sec[i] for i in l] 

        if feedback_function_type: 
            feedback_function = prng_point_distance_funtion(prng)
        else: 
            feedback_function = euclidean_point_distance

        return IsoRing(sec_list,feedback_function,actual_sec_index=actual_sec_index)


from .background_info import * 


def std_cracking_function(ir:IsoRing,hs:HypStruct): 
    assert type(hs) == HypStruct

    ssi = SearchSpaceIterator_for_bounds(hs.suspected_subbound,hs.hop_size)

    """
    empty function 
    """
    def process_feedback(feedback_distance_vec):  
        return

    """
    - cracking point, ?is point correct?,?is finished? 
    """
    def try_one(vec): 
        # calculate feedback vector 
        fvec = ir.provide_feedback_distance_vec(vec)  

        # case: wrong dimension 
        if type(fvec) == type(None): 
            return None,None 

        # no processing of feedback in this cracking function 
        process_feedback(fvec)

        opt_point,opt_index,opt_pr = ir.guess_equals_one_feedback(vec)
        # case: not an optima point 
        if type(opt_point) == type(None): 
            return None,None

        # case: optima point, check if index and pr match 
            # not the index 
        if opt_index != hs.opt_index: 
            return None,None 

            # not probability match
        return opt_point,opt_pr == hs.probability_marker

    def one_guess(): 
        if ssi.reached_end(): 
            return None,None,True 

        v = next(ssi) 
        opt_pt,bool_stat = try_one(v) 
        return opt_pt,bool_stat,False 

    return one_guess 


class Crackling: 

    def __init__(self,target_isoring_idn,isoring_sec_index,sec_optima_index): 
        self.target_ir = target_isoring_idn 
        self.ir_sec_index = isoring_sec_index
        self.sec_opt_index = sec_optima_index
    
        self.num_attempts = 0 
        self.cracked_soln = None
        self.soln_pr = None 

        self.has_bridged = False 
        self.terminated = False 
        return

    def __str__(self): 
        S = "* Target IsoRing: " + str(self.target_ir) + "\n"
        S += "* Sec index: " + str(self.ir_sec_index) + "\n"
        S += "* optima index: " + str(self.sec_opt_index) + "\n"
        S += "* attempts: " + str(self.num_attempts) + "\n"
        S += "* terminated: " + str(self.terminated) + "\n"
        S += "* accept solution: " + str(self.soln_pr) + "\n"
        return S 

    def has_soln(self): 
        return type(self.cracked_soln) != type(None) and \
            type(self.soln_pr) != type(None) 

"""
Cracking bridge. Environment for <Crackling> to attempt cracking an <IsoRing> using a 
<HypStruct>. 
"""
class CBridge:

    def __init__(self,cr:Crackling,hs:HypStruct,ir:IsoRing,cracking_func=std_cracking_function,\
        verbose=False): 
        assert type(cr) == Crackling 
        assert type(hs) == HypStruct 
        assert type(ir) == IsoRing 
        assert type(verbose) == bool 

        self.cr = cr 
        self.hs = hs 
        self.ir = ir 
        self.verbose = verbose 

        self.cracking_process = cracking_func(self.ir,self.hs)
        self.cr.has_bridged = True 
        self.terminated = False 
        return

    def __str__(self): 
        return "CBRIDGE ON ISORING {} WITH HYP\n{}\n".format(self.ir.idn_tag(),str(self.hs)) 

    def __next__(self): 
        if self.terminated: return 0 

        point,bool_stat,fin_stat = self.cracking_process()
        if self.verbose: 
            print("i: ",self.cr.num_attempts)
            print("-------------------------------") 

        if fin_stat: 
            self.terminated = True 

        if type(point) != type(None): 
            self.terminated = True
            self.cr.cracked_soln = point 
            self.cr.soln_pr = bool_stat 
            self.cr.terminated = True 
            if self.verbose: print("[got point] terminated @",self.cr.num_attempts)

        else: 
            if fin_stat: 
                #self.terminated = True 
                self.cr.terminated = True 
                if self.verbose: print("[no point] terminated @",self.cr.num_attempts)

        self.cr.num_attempts += 1
        return 1 